@@ -30,6 +30,7 @@
 #include "cfg.h"
 #include "atomic.h"
 #include "messages.h"
+#include "signal-slot-connector/signal-slot-connector.h"
 
 /* notify code values */
 #define NC_CLOSE       1
@@ -45,23 +46,30 @@
 /* indicates that this LogPipe got cloned into the tree already */
 #define PIF_INLINED           0x0002
 
-/* log statement flags that are copied to the head of a branch */
-#define PIF_BRANCH_FINAL      0x0004
-#define PIF_BRANCH_FALLBACK   0x0008
-#define PIF_BRANCH_PROPERTIES (PIF_BRANCH_FINAL + PIF_BRANCH_FALLBACK)
-
-#define PIF_DROP_UNMATCHED    0x0010
-
-/* branch starting with this pipe wants hard flow control */
-#define PIF_HARD_FLOW_CONTROL 0x0020
-
 /* this pipe is a source for messages, it is not meant to be used to
  * forward messages, syslog-ng will only use these pipes for the
  * left-hand side of the processing graph, e.g. no other pipes may be
  * sending messages to these pipes and these are expected to generate
  * messages "automatically". */
 
-#define PIF_SOURCE            0x0040
+#define PIF_SOURCE            0x0004
+
+/* log statement flags that are copied to the head of a branch */
+#define PIF_BRANCH_FINAL      0x0008
+#define PIF_BRANCH_FALLBACK   0x0010
+#define PIF_BRANCH_PROPERTIES (PIF_BRANCH_FINAL + PIF_BRANCH_FALLBACK)
+
+/* branch starting with this pipe wants hard flow control */
+#define PIF_HARD_FLOW_CONTROL 0x0020
+
+/* LogPipe right after the filter in an "if (filter)" expression */
+#define PIF_CONDITIONAL_MIDPOINT  0x0040
+
+/* LogPipe as the joining element of a junction */
+#define PIF_JUNCTION_END          0x0080
+
+/* node created directly by the user */
+#define PIF_CONFIG_RELATED    0x0100
 
 /* private flags range, to be used by other LogPipe instances for their own purposes */
 
@@ -206,10 +214,50 @@ struct _LogPathOptions
   gboolean flow_control_requested;
 
   gboolean *matched;
+  const LogPathOptions *parent;
 };
 
-#define LOG_PATH_OPTIONS_INIT { TRUE, FALSE, NULL }
-#define LOG_PATH_OPTIONS_INIT_NOACK { FALSE, FALSE, NULL }
+#define LOG_PATH_OPTIONS_INIT { TRUE, FALSE, NULL, NULL }
+#define LOG_PATH_OPTIONS_INIT_NOACK { FALSE, FALSE, NULL, NULL }
+
+static inline void
+log_path_options_push_junction(LogPathOptions *local_path_options, gboolean *matched, const LogPathOptions *parent)
+{
+  *local_path_options = *parent;
+  local_path_options->matched = matched;
+  local_path_options->parent = parent;
+}
+
+static inline void
+log_path_options_pop_conditional(LogPathOptions *local_path_options)
+{
+  if (local_path_options->parent)
+    local_path_options->matched = local_path_options->parent->matched;
+}
+
+/*
+ * NOTE: we need to be optional about ->parent being set, as synthetic
+ * messages (e.g.  the likes emitted by db-parser/grouping-by() may arrive
+ * at the end of a junction without actually crossing the beginning of the
+ * same junction.  But this is ok, in these cases we don't need to propagate
+ * our matched state to anywhere, we can assume that the synthetic message
+ * will just follow the same route as the one it was created from.
+ */
+static inline void
+log_path_options_pop_junction(LogPathOptions *local_path_options)
+{
+  log_path_options_pop_conditional(local_path_options);
+
+  if (local_path_options->parent)
+    local_path_options->parent = local_path_options->parent->parent;
+}
+
+typedef struct _LogPipeOptions LogPipeOptions;
+
+struct _LogPipeOptions
+{
+  gboolean internal;
+};
 
 struct _LogPipe
 {
@@ -224,13 +272,23 @@ struct _LogPipe
   StatsCounterItem *discarded_messages;
   const gchar *persist_name;
   gchar *plugin_name;
+  SignalSlotConnector *signal_slot_connector;
+  LogPipeOptions options;
 
   gboolean (*pre_init)(LogPipe *self);
   gboolean (*init)(LogPipe *self);
   gboolean (*deinit)(LogPipe *self);
   void (*post_deinit)(LogPipe *self);
 
+  gboolean (*pre_config_init)(LogPipe *self);
+  /* this event function is used to perform necessary operation, such as
+   * starting worker thread, and etc. therefore, syslog-ng will terminate if
+   * return value is false.
+   */
+  gboolean (*post_config_init)(LogPipe *self);
+
   const gchar *(*generate_persist_name)(const LogPipe *self);
+  GList *(*arcs)(LogPipe *self);
 
   /* clone this pipe when used in multiple locations in the processing
    * pipe-line. If it contains state, it should behave as if it was
@@ -240,6 +298,7 @@ struct _LogPipe
 
   void (*free_fn)(LogPipe *self);
   void (*notify)(LogPipe *self, gint notify_code, gpointer user_data);
+  GList *info;
 };
 
 /*
@@ -255,9 +314,10 @@ G_STATIC_ASSERT(G_STRUCT_OFFSET(LogPipe, queue) - G_STRUCT_OFFSET(LogPipe, flags
 extern gboolean (*pipe_single_step_hook)(LogPipe *pipe, LogMessage *msg, const LogPathOptions *path_options);
 
 LogPipe *log_pipe_ref(LogPipe *self);
-void log_pipe_unref(LogPipe *self);
+gboolean log_pipe_unref(LogPipe *self);
 LogPipe *log_pipe_new(GlobalConfig *cfg);
 void log_pipe_init_instance(LogPipe *self, GlobalConfig *cfg);
+void log_pipe_clone_method(LogPipe *dst, const LogPipe *src);
 void log_pipe_forward_notify(LogPipe *self, gint notify_code, gpointer user_data);
 EVTTAG *log_pipe_location_tag(LogPipe *pipe);
 void log_pipe_attach_expr_node(LogPipe *self, LogExprNode *expr_node);
@@ -266,6 +326,7 @@ void log_pipe_detach_expr_node(LogPipe *self);
 static inline GlobalConfig *
 log_pipe_get_config(LogPipe *s)
 {
+  g_assert(s->cfg);
   return s->cfg;
 }
 
@@ -316,6 +377,22 @@ log_pipe_deinit(LogPipe *s)
   return TRUE;
 }
 
+static inline gboolean
+log_pipe_pre_config_init(LogPipe *s)
+{
+  if (s->pre_config_init)
+    return s->pre_config_init(s);
+  return TRUE;
+}
+
+static inline gboolean
+log_pipe_post_config_init(LogPipe *s)
+{
+  if (s->post_config_init)
+    return s->post_config_init(s);
+  return TRUE;
+}
+
 static inline void
 log_pipe_queue(LogPipe *s, LogMessage *msg, const LogPathOptions *path_options);
 
@@ -335,6 +412,7 @@ log_pipe_forward_msg(LogPipe *self, LogMessage *msg, const LogPathOptions *path_
 static inline void
 log_pipe_queue(LogPipe *s, LogMessage *msg, const LogPathOptions *path_options)
 {
+  LogPathOptions local_path_options;
   g_assert((s->flags & PIF_INITIALIZED) != 0);
 
   if (G_UNLIKELY(pipe_single_step_hook))
@@ -346,14 +424,23 @@ log_pipe_queue(LogPipe *s, LogMessage *msg, const LogPathOptions *path_options)
         }
     }
 
-  if (G_UNLIKELY(s->flags & (PIF_HARD_FLOW_CONTROL)))
+  if (G_UNLIKELY(s->flags & (PIF_HARD_FLOW_CONTROL | PIF_JUNCTION_END | PIF_CONDITIONAL_MIDPOINT)))
     {
-      LogPathOptions local_path_options = *path_options;
-
-      local_path_options.flow_control_requested = 1;
+      local_path_options = *path_options;
+      if (s->flags & PIF_HARD_FLOW_CONTROL)
+        {
+          local_path_options.flow_control_requested = 1;
+          msg_trace("Requesting flow control", log_pipe_location_tag(s));
+        }
+      if (s->flags & PIF_JUNCTION_END)
+        {
+          log_path_options_pop_junction(&local_path_options);
+        }
+      if (s->flags & PIF_CONDITIONAL_MIDPOINT)
+        {
+          log_path_options_pop_conditional(&local_path_options);
+        }
       path_options = &local_path_options;
-
-      msg_trace("Requesting flow control", log_pipe_location_tag(s));
     }
 
   if (s->queue)
@@ -365,10 +452,6 @@ log_pipe_queue(LogPipe *s, LogMessage *msg, const LogPathOptions *path_options)
       log_pipe_forward_msg(s, msg, path_options);
     }
 
-  if (path_options->matched && !(*path_options->matched) && (s->flags & PIF_DROP_UNMATCHED))
-    {
-      (*path_options->matched) = TRUE;
-    }
 }
 
 static inline LogPipe *
@@ -397,6 +480,17 @@ log_pipe_set_persist_name(LogPipe *self, const gchar *persist_name);
 const gchar *
 log_pipe_get_persist_name(const LogPipe *self);
 
+void log_pipe_set_options(LogPipe *self, const LogPipeOptions *options);
+void log_pipe_set_internal(LogPipe *self, gboolean internal);
+gboolean log_pipe_is_internal(const LogPipe *self);
+
 void log_pipe_free_method(LogPipe *s);
 
+static inline GList *
+log_pipe_get_arcs(LogPipe *s)
+{
+  return s->arcs(s);
+}
+
+void log_pipe_add_info(LogPipe *self, const gchar *info);
 #endif
